# This file is part of tad-dftd3.
# SPDX-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dispersion model
================

Implementation of D3 model to obtain atomic C6 coefficients for a given geometry.

Examples
--------
>>> import torch
>>> import tad_dftd3 as d3
>>> import tad_mctc as mctc
>>> numbers = mctc.convert.symbol_to_number(["O", "H", "H"])
>>> positions = torch.Tensor([
...     [+0.00000000000000, +0.00000000000000, -0.73578586109551],
...     [+1.44183152868459, +0.00000000000000, +0.36789293054775],
...     [-1.44183152868459, +0.00000000000000, +0.36789293054775],
... ])
>>> ref = d3.reference.Reference()
>>> rcov = d3.data.covalent_rad_d3[numbers]
>>> cn = mctc.ncoord.cn_d3(numbers, positions, rcov=rcov, counting_function=d3.ncoord.exp_count)
>>> weights = d3.model.weight_references(numbers, cn, ref, d3.model.gaussian_weight)
>>> c6 = d3.model.atomic_c6(numbers, weights, ref)
>>> torch.set_printoptions(precision=7)
>>> print(c6)
tensor([[10.4130471,  5.4368822,  5.4368822],
        [ 5.4368822,  3.0930154,  3.0930154],
        [ 5.4368822,  3.0930154,  3.0930154]], dtype=torch.float64)
"""
from __future__ import annotations

import torch
from tad_mctc import storch

from ..reference import Reference
from ..typing import Any, Tensor, WeightingFunction

__all__ = ["gaussian_weight", "weight_references"]


def gaussian_weight(dcn: Tensor, factor: float = 4.0) -> Tensor:
    """
    Calculate weight of indivdual reference system.

    Parameters
    ----------
    dcn : Tensor
        Difference of coordination numbers.
    factor : float
        Factor to calculate weight.

    Returns
    -------
    Tensor
        Weight of individual reference system.
    """

    return torch.exp(-factor * dcn.pow(2))


def weight_references(
    numbers: Tensor,
    cn: Tensor,
    reference: Reference,
    weighting_function: WeightingFunction = gaussian_weight,
    **kwargs: Any,
) -> Tensor:
    """
    Calculate the weights of the reference system.

    Parameters
    ----------
    numbers : Tensor
        The atomic numbers of the atoms in the system.
    cn : Tensor
        Coordination numbers for all atoms in the system.
    reference : Reference
        Reference systems for D3 model.
    weighting_function : Callable
        Function to calculate weight of individual reference systems.

    Returns
    -------
    Tensor
        Weights of all reference systems
    """
    refcn = reference.cn[numbers]
    mask = refcn >= 0

    zero = torch.tensor(0.0, device=cn.device, dtype=cn.dtype)
    zero_double = torch.tensor(0.0, device=cn.device, dtype=torch.double)
    one = torch.tensor(1.0, device=cn.device, dtype=cn.dtype)

    # Due to the exponentiation, `norms` and `weights` may become very small.
    # This may cause problems for the division by `norms`. It may occur that
    # `weights` and `norms` are equal, in which case the result should be
    # exactly one. This might, however, not be the case and ultimately cause
    # larger deviations in the final values.
    #
    # This must be done in the D4 variant because the weighting functions
    # contains higher powers, which lead to values down to 1e-300.
    # Since there are also cases in D3, we have to evaluate this portion
    # in double precision to retain the correct results and avoid nan's.
    dcn = (reference.cn[numbers] - cn.unsqueeze(-1)).type(torch.double)
    weights = torch.where(
        mask,
        weighting_function(dcn, **kwargs),
        zero_double,  # not eps!
    )

    # Previously, a small value was added to `norms` to prevent division by zero
    # (`norms = torch.add(torch.sum(weights, dim=-1), 1e-20)`). However, even
    # such small values can lead to relatively large deviations because the
    # small value is not added to the weights, and hence, the case where
    # `weights` and `norms` are equal does not yield one anymore. In fact, the
    # test suite fails because some elements deviate up to around 1e-4.
    # We solve this by running in double precision, adding a very small number
    # and using multiple masks.

    small = torch.tensor(1e-300, device=cn.device, dtype=torch.double)

    # normalize weights
    norm = torch.where(
        mask,
        torch.sum(weights, dim=-1, keepdim=True),
        small,  # double!
    )

    # back to real dtype
    gw_temp = storch.divide(weights, norm, eps=small).type(cn.dtype)
    assert torch.isnan(gw_temp).sum() == 0

    # The following section handles cases with large CNs that lead to zeros in
    # after the exponential in the weighting function. If this happens all
    # weights become zero, which is not desired. Instead, we set the weight of
    # the largest reference number to one.
    # This case can occur if the CN of the current (actual) system is too far
    # away from the largest CN of the reference systems. An example would be an
    # atom within a fullerene (La3N@C80).

    # maximum reference CN for each atom
    maxcn = torch.max(refcn, dim=-1, keepdim=True)[0]

    # Here, we catch the potential NaN's from `gw_temp`. We cannot use `gw_temp`
    # directly, because we have to use safe divide to not get NaN's in the
    # backward. But `norm == 0` is equivalent. Additionally, we catch very
    # large values occuring because of division by small values.
    exceptional = (norm == 0) | (gw_temp > torch.finfo(cn.dtype).max)

    gw = torch.where(
        exceptional,
        torch.where(refcn == maxcn, one, zero),
        gw_temp,
    )

    return torch.where(mask, gw, zero)



# This file is part of tad-dftd3.
# SPDX-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model: Atomic C6
================

Computation of atomic C6 dispersion coefficients.

Since this part can be the most memory-intensive, we provide a custom backward
function (i.e., analytical gradient) and options for chunking.
"""
from __future__ import annotations

import torch
from tad_mctc._version import __tversion__
from tad_mctc.math import einsum
from tad_mctc.tools import memory

from ..reference import Reference
from ..typing import Callable, Protocol, Tensor

__all__ = ["atomic_c6"]


# main entry point


def atomic_c6(
    numbers: Tensor,
    weights: Tensor,
    reference: Reference,
    chunk_size: None | int = None,
) -> Tensor:
    """
    Calculate atomic dispersion coefficients.

    .. warning::

        This function is the most memory intensive part of the calculation and
        may require chunking for large systems. Without chunking, for example,
        2000 atoms (`numbers`) require the construction of a 1.5 GB tensor.

    Parameters
    ----------
    numbers : Tensor
        The atomic numbers of the atoms in the system of shape `(..., nat)`.
    weights : Tensor
        Weights of all reference systems of shape `(..., nat, 7)`.
    reference : Reference
        Reference systems for D3 model. Contains the reference C6 coefficients
        of shape `(..., nelements, nelements, 7, 7)`.

    Returns
    -------
    Tensor
        Atomic dispersion coefficients of shape `(..., nat, nat)`.
    """
    _check_memory(numbers, weights, chunk_size)

    # PyTorch 2.0.x has a bug with functorch and custom autograd functions as
    # documented in: https://github.com/pytorch/pytorch/issues/99973
    #
    # RuntimeError: unwrapped_count > 0 INTERNAL ASSERT FAILED at "../aten/src/
    # ATen/functorch/TensorWrapper.cpp":202, please report a bug to PyTorch.
    # Should have at least one dead wrapper
    #
    # Hence, we cannot use the custom backwards for reduced memory consumption.
    if __tversion__[0] == 2 and __tversion__[1] == 0:  # pragma: no cover
        track_weights = torch._C._functorch.is_gradtrackingtensor(weights)
        track_numbers = torch._C._functorch.is_gradtrackingtensor(numbers)
        if track_weights or track_numbers:

            if chunk_size is None:
                return _atomic_c6_full(numbers, weights, reference)

            return _atomic_c6_chunked(numbers, weights, reference, chunk_size)

    # Use custom autograd function for reduced memory consumption
    AtomicC6 = AtomicC6_V1 if __tversion__ < (2, 0, 0) else AtomicC6_V2
    res = AtomicC6.apply(numbers, weights, reference, chunk_size)
    assert res is not None
    return res


# helpers


def _check_memory(
    numbers: Tensor, weights: Tensor, chunk_size: None | int = None
) -> None:
    """
    Check memory usage for the construction of the C6 tensor.
    Throw an error or warning for potential memory issues.

    Parameters
    ----------
    numbers : Tensor
        Atomic numbers of the atoms in the system.
    weights : Tensor
        Weights of all reference systems.
    chunk_size : None | int, optional
        Chunk size for the calculation of the C6 tensor. Defaults to `None`.

    Raises
    ------
    MemoryError
        If the estimated memory usage exceeds the total available memory.
    """
    # Required memory for the C6 tensor
    if chunk_size is None:
        size = (numbers.shape[-1], numbers.shape[-1], 7, 7)
    else:
        size = (numbers.shape[-1], chunk_size, 7, 7)
    mem = memory.memory_tensor(size, weights.dtype)

    # actual memory usage
    free, total = memory.memory_device(numbers.device)

    if mem > total:
        raise MemoryError(
            f"Estimated memory usage exceeds total available memory: {mem:.2f} "
            f"MB > {total:.2f} MB. During the construction of the C6 "
            f"dispersion coefficients, a 4D tensor of shape {size} is required "
            "for efficient tensor operations. To fit the tensor into memory, "
            "try using a chunk size or reduce the chunk size via the optional "
            "`chunk_size` argument."
        )

    if mem > free:
        # pylint: disable=import-outside-toplevel
        from warnings import warn

        warn(
            "Estimated memory usage appears to exceed the available memory: "
            f"{mem:.2f} MB > {free:.2f} MB. If the calculation fails due to "
            "memory issues, consider reducing the chunk size via the optional "
            "`chunk_size` argument.",
            ResourceWarning,
        )


def _einsum(rc6: Tensor, weights_i: Tensor, weights_j: Tensor) -> Tensor:
    """
    Perform an einsum operation for the atomic C6 coefficients.

    Parameters
    ----------
    rc6 : Tensor
        Reference C6 coefficients.
    weights_i : Tensor
        Weights of all reference systems.
    weights_j : Tensor
        Weights of all reference systems.

    Returns
    -------
    Tensor
        Atomic C6 dispersion coefficients.
    """
    # The default einsum path is fastest if the large tensors comes first.
    # (..., n1, n2, r1, r2) * (..., n1, r1) * (..., n2, r2) -> (..., n1, n2)
    return einsum(
        "...ijab,...ia,...jb->...ij",
        *(rc6, weights_i, weights_j),
        optimize=[(0, 1), (0, 1)],
    )


# full and chunked versions


def _atomic_c6_full(
    numbers: Tensor,
    weights: Tensor,
    reference: Reference,
) -> Tensor:
    """
    Calculation of atomic dispersion coefficients without chunking. Might cause
    memory issues for very large systems.

    Parameters
    ----------
    numbers : Tensor
        The atomic numbers of the atoms in the system of shape `(..., nat)`.
    weights : Tensor
        Weights of all reference systems of shape `(..., nat, 7)`.
    reference : Reference
        Reference systems for D3 model. Contains the reference C6 coefficients
        of shape `(..., nelements, nelements, 7, 7)`.

    Returns
    -------
    Tensor
        Atomic dispersion coefficients of shape `(..., nat, nat)`.
    """
    # NOTE: This old version creates large intermediate tensors and builds the
    # full matrix before the sum reduction, which requires a lot of memory.
    #
    # gw = w.unsqueeze(-1).unsqueeze(-3) * w.unsqueeze(-2).unsqueeze(-4)
    # c6 = torch.sum(torch.sum(torch.mul(gw, rc6), dim=-1), dim=-1)

    rc6 = reference.c6[numbers.unsqueeze(-1), numbers.unsqueeze(-2)]
    return _einsum(rc6, weights, weights)


def _atomic_c6_chunked(
    numbers: Tensor,
    weights: Tensor,
    reference: Reference,
    chunk_size: int,
) -> Tensor:
    """
    Chunked version of the calculation of atomic dispersion coefficients.

    Parameters
    ----------
    numbers : Tensor
        The atomic numbers of the atoms in the system of shape `(..., nat)`.
    weights : Tensor
        Weights of all reference systems of shape `(..., nat, 7)`.
    reference : Reference
        Reference systems for D3 model. Contains the reference C6 coefficients
        of shape `(..., nelements, nelements, 7, 7)`.
    chunk_size : int
        Chunk size for the calculation of the C6 tensor.

    Returns
    -------
    Tensor
        Atomic dispersion coefficients of shape `(..., nat, nat)`.
    """

    nat = numbers.shape[-1]
    c6_output = torch.zeros(
        (*numbers.shape, nat), device=numbers.device, dtype=weights.dtype
    )

    for start in range(0, nat, chunk_size):
        end = min(start + chunk_size, nat)
        num_chunk = numbers[..., start:end]  # (..., chunk_size)

        # Chunked indexing into reference.c6: (..., chunk_size, nat, 7, 7)
        rc6_chunk = reference.c6[num_chunk.unsqueeze(-1), numbers.unsqueeze(-2)]

        # Also chunk the weights: (..., chunk_size, 7)
        weights_chunk = weights[..., start:end, :]

        # (..., n1, n2, r1, r2) * (..., n1, r1) * (..., n2, r2) -> (..., n1, n2)
        contribution = _einsum(rc6_chunk, weights_chunk, weights)

        # Add contributions to the correct slice of the output tensor
        c6_output[..., start:end, :] += contribution

    return c6_output


# custom autograd functions


class CTX(Protocol):
    save_for_backward: Callable[[Tensor, Tensor], None]
    saved_tensors: tuple[Tensor, Tensor]
    chunk_size: None | int
    reference: Reference


class AtomicC6Base(torch.autograd.Function):
    """
    Base class for the version-specific autograd function for atomic C6.
    Different PyTorch versions only require different `forward()` signatures.
    """

    @staticmethod
    def backward(ctx: CTX, grad_out: Tensor) -> tuple[None, Tensor, None, None]:
        numbers, weights = ctx.saved_tensors
        chunk_size = ctx.chunk_size
        ref = ctx.reference

        # We need the derivatives of the following expression:
        # c_ij ​= ∑a,b w_ia *× w_jb ​* c_ijab​

        ###########################
        ### Non-chunked version ###
        ###########################

        if chunk_size is None:
            # (..., nel, nel, 7, 7) -> (..., nat, nat, 7, 7)
            rc6 = ref.c6[numbers.unsqueeze(-1), numbers.unsqueeze(-2)]

            # ∂c_ij/∂w_jb = ∑a w_ia * c_ijab
            # (..., n1, n2, r1, r2) * (..., n2, r2) -> (..., n1, n2, r2)
            g_jb = einsum("...ijab,...ia->...ijb", rc6, weights)

            # vjp: (..., n1, n2) * (..., n1, n2, r2) -> (..., n2, r2)
            _gj = einsum("...ij,...ijb->...jb", grad_out, g_jb)

            # ∂c_ij/∂w_ia = ∑b w_jb * c_ijab
            # (..., n1, n2, r1, r2) * (..., n2, r2) -> (..., n1, n2, r1)
            g_ia = einsum("...ijab,...jb->...ija", rc6, weights)

            # vjp: (..., n1, n2) * (..., n1, n2, r1) -> (..., n1, r1)
            _gi = einsum("...ij,...ija->...ia", grad_out, g_ia)

            weights_bar = _gi + _gj

            return None, weights_bar, None, None

        #######################
        ### Chunked version ###
        #######################

        nat = weights.shape[-2]
        weights_bar = torch.zeros_like(weights)

        for start in range(0, nat, chunk_size):
            end = min(start + chunk_size, nat)

            # Numbers and derivatives for this chunk
            grad_chunk = grad_out[..., start:end, :]  # (..., chunk_size, nat)
            num_chunk = numbers[..., start:end]  # (..., chunk_size)

            # Chunked indexing into reference.c6: (..., chunk_size, nat, 7, 7)
            # -> Only the "i" index is chunked!
            rc6_chunk = ref.c6[num_chunk.unsqueeze(-1), numbers.unsqueeze(-2)]

            # Also chunk the weights: (..., chunk_size, 7)
            weights_chunk = weights[..., start:end, :]

            # _gi derivative is chunked (sum over non-chunked "j" index)
            g_ia = einsum("...ijab,...jb->...ija", rc6_chunk, weights)
            _gi = einsum("...ij,...ija->...ia", grad_chunk, g_ia)

            # _gj derivative is NOT chunked (sum over chunked "i" index)
            g_jb = einsum("...ijab,...ia->...ijb", rc6_chunk, weights_chunk)
            _gj = einsum("...ij,...ijb->...jb", grad_chunk, g_jb)

            # Accumulate gradients for the current chunk
            weights_bar[..., start:end, :] += _gi
            weights_bar += _gj

        return None, weights_bar, None, None


class AtomicC6_V1(AtomicC6Base):
    """
    Custom autograd function for atomic C6 coefficients.
    This is supposed to reduce memory usage.
    """

    @staticmethod
    def forward(
        ctx: CTX,
        numbers: Tensor,
        weights: Tensor,
        reference: Reference,
        chunk_size: None | int = None,
    ) -> Tensor:
        ctx.save_for_backward(numbers, weights)
        ctx.chunk_size = chunk_size
        ctx.reference = reference

        if chunk_size is None:
            return _atomic_c6_full(numbers, weights, reference)

        return _atomic_c6_chunked(numbers, weights, reference, chunk_size)


class AtomicC6_V2(AtomicC6Base):
    """
    Custom autograd function for atomic C6 coefficients.
    This is supposed to reduce memory usage.
    """

    generate_vmap_rule = True
    # https://pytorch.org/docs/master/notes/extending.func.html#automatically-generate-a-vmap-rule
    # should work since we only use PyTorch operations

    @staticmethod
    def forward(
        numbers: Tensor,
        weights: Tensor,
        reference: Reference,
        chunk_size: None | int = None,
    ) -> Tensor:
        if chunk_size is None:
            return _atomic_c6_full(numbers, weights, reference)

        return _atomic_c6_chunked(numbers, weights, reference, chunk_size)

    @staticmethod
    def setup_context(
        ctx: CTX,
        inputs: tuple[Tensor, Tensor, Reference, int | None],
        output: Tensor,
    ) -> None:
        numbers, weights, reference, chunk_size = inputs

        ctx.save_for_backward(numbers, weights)
        ctx.chunk_size = chunk_size
        ctx.reference = reference



# This file is part of tad-dftd3.
# SPDX-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defaults
========

This module defines the default values for all parameters within DFT-D3.
"""

__all__ = [
    "D3_CN_CUTOFF",
    "D3_DISP_CUTOFF",
    "D3_KCN",
    "A1",
    "A2",
    "S6",
    "S8",
    "S9",
    "RS9",
    "ALP",
    "MAX_ELEMENT",
]

# DFT-D3

D3_CN_CUTOFF = 25.0
"""Coordination number cutoff (25.0)."""

D3_DISP_CUTOFF = 50.0
"""Two/three-body interaction cutoff (50.0)."""

D3_KCN = 16.0
"""Steepness of counting function (16.0)."""

# DFT-D3 damping parameters

A1 = 0.4
"""Scaling for the C8 / C6 ratio in the critical radius (0.4)."""

A2 = 5.0
"""Offset parameter for the critical radius (5.0)."""

S6 = 1.0
"""Scaling factor for the C6 interaction (1.0)."""

S8 = 1.0
"""Scaling factor for the C8 interaction (1.0)."""

S9 = 1.0
"""Scaling for dispersion coefficients (1.0)."""

RS9 = 4.0 / 3.0
"""Scaling for van-der-Waals radii in damping function (4.0/3.0)."""

ALP = 14.0
"""Exponent of zero damping function (14.0)."""

# other

MAX_ELEMENT = 104
"""Atomic number (+1 for dummy) of last element supported by DFT-D3."""



# This file is part of tad-dftd3.
# SPDX-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model: Dispersion model
=======================

Implementation of D3 model to obtain atomic C6 coefficients for a given
geometry.

Examples
--------
>>> import torch
>>> import tad_dftd3 as d3
>>> import tad_mctc as mctc
>>> numbers = mctc.convert.symbol_to_number(["O", "H", "H"])
>>> positions = torch.Tensor([
...     [+0.00000000000000, +0.00000000000000, -0.73578586109551],
...     [+1.44183152868459, +0.00000000000000, +0.36789293054775],
...     [-1.44183152868459, +0.00000000000000, +0.36789293054775],
... ])
>>> ref = d3.reference.Reference()
>>> rcov = d3.data.covalent_rad_d3[numbers]
>>> cn = mctc.ncoord.cn_d3(numbers, positions, rcov=rcov, counting_function=d3.ncoord.exp_count)
>>> weights = d3.model.weight_references(numbers, cn, ref, d3.model.gaussian_weight)
>>> c6 = d3.model.atomic_c6(numbers, weights, ref)
>>> torch.set_printoptions(precision=7)
>>> print(c6)
tensor([[10.4130471,  5.4368822,  5.4368822],
        [ 5.4368822,  3.0930154,  3.0930154],
        [ 5.4368822,  3.0930154,  3.0930154]], dtype=torch.float64)
"""
from .c6 import *
from .weights import *



# This file is part of tad-dftd3.
# SPDX-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data: Radii
===========

Data arrays for atomic constants like covalent radii or van-der-Waals radii.

The `vdw_rad_d3` were previously stored explicitly in one list and then
reshaped to the required `(MAX_ELEMENT, MAX_ELEMENT)` tensor. For the old
version, see older commits (e.g. https://github.com/dftd3/tad-dftd3/blob/ecc50f19adb8aa8baa38a188d04228c4f26975d6/src/tad_dftd3/data/radii.py)
"""
import os.path as op
from typing import Optional

import torch
from tad_mctc.data.radii import COV_D3

from ..typing import Tensor

__all__ = ["COV_D3", "VDW_D3"]


def _load_vdw_rad_d3(
    dtype: torch.dtype = torch.double, device: Optional[torch.device] = None
) -> Tensor:
    """
    Load reference VDW radii from file.

    Parameters
    ----------
    dtype : torch.dtype, optional
        Floating point precision for tensor. Defaults to `torch.double`.
    device : Optional[torch.device], optional
        Device of tensor. Defaults to None.

    Returns
    -------
    Tensor
        VDW radii.
    """
    path = op.join(op.dirname(__file__), "vdw-d3.pt")
    return torch.load(path).type(dtype).to(device)


VDW_D3 = _load_vdw_rad_d3()
